#pragma once
#include "stdafx.h"

#include <boost/fusion/algorithm/iteration/for_each.hpp> 
#include <boost/fusion/algorithm/iteration/fold.hpp>
#include <boost/fusion/container/generation/vector_tie.hpp>
#include <boost/fusion/functional/invocation/invoke.hpp>  
#include <boost/fusion/functional/invocation/invoke_procedure.hpp> 
#include <boost/numeric/conversion/cast.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>

#include "PG_LibPQTypes.h"

#include "winlogger.h"
#include "_project.h"
#include "exceptions.h"
#include "tbl_cmp_predic.h"

#include "CDBExecutorUtility.h"
#include "smartpqtypesutils.h"

namespace fusion = boost::fusion;

#define	DEALLOCATE_PREPARE "DEALLOCATE"
#define	PREP_PREFIX "PREPSTMT"

// log connection string as well
#define _DO_SHOW_CONNECTION_DETAILS 1 

#define ERROR_HANDLER_START() \
	try\
{

#define ERROR_HANDLER_END() \
}\
	catch ( Smartdb::smartdb_exception& ex )\
{\
	if (!errorHandler_)\
{\
	throw;\
}\
	errorHandler_(ex);\
}
typedef void (*SmartDBErrHandlerFunc)(const Smartdb::smartdb_exception &);

/**
*  The class provides the base interactions with Postgres DB, designed using Singleton Design Pattern.
*  The base purpose of the class is to provide a complete set of functions for interaction with Postgres DB.
*/
class PG_LIBPQTYPES_API CDBExecutor
{
public:
	typedef std::map<std::string, std::string> TPrepCmdRep;
	/**
	* Constructor.
	* @param connMan pointer to a connection manager
	*/
	CDBExecutor(ConnectionManager *connMan, WinLogger *logger); 
private:
	//static CDBExecutor* instance_; 

	/** table instance */
	//boost::variant<Inst::inst_*, Asset_spot_combination::asset_spot_combination_*> tbl_;
	/** Connection to PG database */
	PGconn* conn_;
	std::string stmt_;
	bool is_composite_;
	boost::scoped_ptr<Detail::DBManager::QueryParams> params_;
	boost::scoped_ptr<Detail::DBManager> pg_mng_;
	std::string tablename_; 
	std::string comp_type_; 
	ConnectionManager *connManager_;
#pragma warning(push)
#pragma warning(disable:4251)
	TPrepCmdRep prepCmdRep_;
#pragma warning(pop)
	std::string defStmtName_;
	/**
	* Logger.
	* Used for logging {fatal, error, warning, info, debug, trace} info
	*/
	WinLogger* logger_;
	/**
	* Error handler function.
	* To extract point of error use: "getErrPoint" smartdb_exception's function
	* To get error message use: "base().what()" smartdb_exception's functions
	* for casting to a type of error: "base().what()" smartdb_exception's functions
	* Example:
	* \code
	* void errHandlerFunc(const pqxx::pqxx_exception &e)
	* {
	*   std::cerr << e.getErrPoint() << << e.base().what() << std::endl;
	*   //for casting to a type of error
	*   const pqxx::sql_error *s=dynamic_cast<const pqxx::sql_error*>(&e.base());
	*   if (s) std::cerr << "Query was: " << s->query() << std::endl;
	* \endcode
	*/
	SmartDBErrHandlerFunc errorHandler_;
private:
	std::string getStatement(void) const;
	void setStatement(const std::string&);
	void PQExecSQL(const std::string &cmd, PGparam* param, const std::string &errMsg, int format = 1);//format: 0 - text, 1 - binary
	void PQExecPreparedSQL(const std::string &cmdName, PGparam* param, const std::string &errMsg, int format);
	bool removeFromPrepCmdRep(/*IN*/const std::string &stmtName);
	void cleanUp(void);
	void prepareConnection();
	//void doInsert() throw(...);
	//void doInsertPrepared(const std::string &stmtName, int format) throw(...);
	void adjustSQLCmd(const std::string &in, std::string &out);
	template <class V> 
	void addParam(const char *paramType, const V* value, const boost::false_type&);
	void addParam(const char *paramType, const std::string* value, const boost::false_type&);
	template <class V> 
	void addParam(const char *paramType, const V* value, const boost::true_type&);

public:
	/**
	* Creates a prepared command for later execution.
	* @param sqlCmd SQL command
	* @return generated statement name
	*/
	std::string prepareAutoNameSQL(const std::string &sqlCmd);
	/**
	* Creates a prepared command for later execution.
	* The function creates prepared named statement.Any pre-prepared named command from the current session with the same name is automatically replaced.
	* @param sqlCmd SQL command
	* @param stmtName name of the prepared SQL command
	* Example:
	* \code
	* prepareSQL("SELECT * FORM tbl", "select12"); \endcode
	*/
	void prepareSQL(const std::string &sqlCmd, const std::string &stmtName);
	/**
	* Executes SQL command.
	* @param cmd SQL command
	* @param format desired format: '0' - text, '1' - binary
	*/
	//void execSQL(const std::string &cmd, int format = 1);
	void execSQL(const std::string &cmd, int format = 1);
	/**
	* Executes SQL command.
	* The only difference between this function and execPreparedSQL(int) is that 
	* this function expects a statement name stmtName 
	* @param stmtName name of the prepared SQL command
	* @remarks if the connection was lost all prepared SQl commands are lost
	*/
	void execPreparedSQL(const std::string &stmtName, int format = 1);
	/**
	* Executes prepared SQL command, storing result data into a table container
	* @param table table container
	* @param stmtName name of the prepared SQL command
	* @param format desired format: '0' - text, '1' - binary
	* @remarks if the connection was lost all prepared SQl commands are lost
	* Example:
	* \code
	*	dbexec->execPreparedSQLGetLast(tbInst);
	*	\endcode
	*/
	template<class T>
	void execPreparedSQLGet(T &tbl, const std::string &stmtName, int format = 1);
	/** Executes prepared SQL command, using data from a table container.
	* Number of the parameters in prepared SQL command must be the same with number of table container's columns,
	* if it is different an exception will throw.
	* If some parameters was added by addParam and not eat used by execPreparedSQL or execSQL commands an exception will throw too,
	* use clearParams() in this case.
	* @param table table container
	* @param format desired format: '0' - text, '1' - binary
	* @param stmtName name of the prepared SQL command
	* @remarks if the connection was lost all prepared SQl commands are lost
	* Example:
	* \code
	*	dbexec->execPreparedSQLSet(tbInst, "select12");
	*	\endcode
	*/
	template<class T>
	void execPreparedSQLSet(T &tbl, const std::string &stmtName, int format = 1);

	/** Deallocates prepared SQL command
	* @param stmtName name of the prepared SQL command
	* @remarks if the connection was lost all prepared SQl commands are lost
	*/
	void deallocPreparedSQL(const std::string &stmtName);
	/** Executes SQL command, using data from a table container.
	* Number of the parameters in SQL command must be the same with number of table container's columns,
	* if it is different an exception will throw.
	* If some parameters was added by addParam and not eat used by execPreparedSQL or execSQL commands an exception will throw too,
	* use clearParams() in this case.
	* @param table table container
	* @param format desired format: '0' - text, '1' - binary
	* @remarks if the connection was lost all prepared SQl commands are lost
	* Example:
	* \code
	*	dbexec->execSQLSet(tbInst, "select12");
	*	\endcode
	*/
	template<class T>
	void execSQLSet(T &tbl, const std::string &SQLCmd, int format = 1);

	//	template<class T>
	//	void execSQLSet(T &tbl, const std::string &SQLCmd, int format = 1);
	/**
	* Executes SQL command, storing result data into a table container
	* @param table table container
	* @param format desired format: '0' - text, '1' - binary
	* @remarks 
	* Example:
	* \code
	*	dbexec->execSQLGet(tbInst);
	*	\endcode
	*/
	//template<class T>
	//void execSQLGet(T &tbl, const std::string &SQLCmd, int format = 1);
	template<class T>
	void execSQLGet(T &tbl, const std::string &SQLCmd, int format = 1);
	~CDBExecutor(void);
	/**
	* Links a table container to another table container
	* @param tobj table container link to
	* @param dobj table to be linked
	*/
	template <typename MLtPred, typename SLtPred, typename MSLtGtEqPred, typename Cnt, typename T, typename S>
	void linkRows(T &tobj, S &dobj, Cnt &cnt);

	template <size_t C1, size_t C2, typename Cnt, typename T, typename S>
	inline void linkRows(T &tobj, S &dobj, Cnt &cnt)
	{
		linkRows<lt_pred<typename T::row_type, C1>, lt_pred<S::row_type, C2>, cmp_pred<C1, T::row_type, C2, typename S::row_type>, Cnt, T, S>(tobj, dobj, cnt);
	}
	template <size_t Cl1, size_t Cl2, size_t Cr1, size_t Cr2, typename Cnt, typename T, typename S>
	inline void linkRows(T &tobj, S &dobj, Cnt &cnt)
	{
		linkRows<lt_pred2<typename T::row_type, Cl1, Cl2>, lt_pred2<S::row_type, Cr1, Cr2>, cmp_pred2<Cl1, Cl2, T::row_type, Cr1, Cr2, typename S::row_type>, Cnt, T, S>(tobj, dobj, cnt);
	}
	template <size_t Cl1, size_t Cl2, size_t Cl3, size_t Cr1, size_t Cr2, size_t Cr3, typename Cnt, typename T, typename S>
	inline void linkRows(T &tobj, S &dobj, Cnt &cnt)
	{
		linkRows<lt_pred3<typename T::row_type, Cl1, Cl2, Cl3>, lt_pred3<S::row_type, Cr1, Cr2, Cr3>, cmp_pred3<Cl1, Cl2, Cl3, T::row_type, Cr1, Cr2, Cr3, typename S::row_type>, Cnt, T, S>(tobj, dobj, cnt);
	}
	/**
	* Sort a table container for given columns
	* @param tobj table container to be sorted
	* Example:
	* \code
	* dbExecutor->sortRows<1>(*master_table, CDBExecutor::SORTORDER_DESC); //sorting by 2nd column
	* \endcode
	*/
	typedef enum {SORTORDER_ASC = 0, SORTORDER_DESC} SortOrderType;

	template <size_t C1, typename T>
	inline void sortRows(T &tobj, SortOrderType stype = SORTORDER_ASC)
	{
		stype == SORTORDER_ASC ? sortRows<lt_pred<typename T::row_type, C1>, T>(tobj)
								:sortRows<gt_pred<typename T::row_type, C1>, T>(tobj);
	}

	template <size_t C1, size_t C2, typename T>
	inline void sortRows(T &tobj, SortOrderType stype = SORTORDER_ASC)
	{
		stype == SORTORDER_ASC ? sortRows<lt_pred2<typename T::row_type, C1, C2>, T>(tobj)
								:sortRows<gt_pred2<typename T::row_type, C1, C2>, T>(tobj);
	}
	template <size_t C1, size_t C2, size_t C3, typename T>
	inline void sortRows(T &tobj, SortOrderType stype = SORTORDER_ASC)
	{
		stype == SORTORDER_ASC ? sortRows<lt_pred3<typename T::row_type, C1, C2, C3>, T>(tobj)
								:sortRows<gt_pred3<typename T::row_type, C1, C2, C3>, T>(tobj);
	}
	/**
	* Adding a parameter to the CDBExecutor which will be applied to SQL query
	*
	* @param paramType type of the parameter
	* Currently supported parameters:
	*		PG_INT8
	*		PG_INT4
	*		PG_INT2
	*		PG_FLOAT8
	*		PG_FLOAT4
	*		PG_VARCHAR
	*		PG_TIMESTAMP
	*		PG_MAC_ADDR
	*		PG_INET
	*		PG_DATE
	*		PG_BOOL_ARR
	*		PG_INT2_ARR
	*		PG_INT4_ARR
	*		PG_INT8_ARR
	*		PG_FLOAT4_ARR
	*		PG_FLOAT8_ARR
	*		PG_VARCHAR_ARR
	*		PG_DATE_ARR
	*		PG_TIMESTAMP_ARR
	*		PG_MAC_ADDR_ARR
	*		PG_INET_ARR
	*		PG_BOOL_VEC
	*		PG_INT2_VEC
	*		PG_INT4_VEC
	*		PG_INT8_VEC
	*		PG_FLOAT4_VEC
	*		PG_FLOAT8_VEC
	*		PG_VARCHAR_VEC
	*		PG_DATE_VEC
	*		PG_TIMESTAMP_VEC
	*		PG_MAC_ADDR_VEC
	*		PG_INET_VEC
	* @param value value of the added parameter
	* @see addParamTimeStamp()
	* Example:
	* \code
	* addParam(PG_INT4, 2); \endcode
	*/

	template <class T>
	void addParam(const char *paramType, multi_array<T> &in_array);
	template <class V> 
	void addParam(const char *paramType, const V& value);
	void addParam(const char *paramType);
	void clearParams();
	/**
	* Sets error handler function
	* @param func pointer to callback function
	* @remart to disable error handler function set it to NULL
	* Example:
	* \code
	* setErrorHandler(&sqlErrorLogger);
	* //smardb calls
	* setErrorHandler(NULL);
	* \endcode
	*/
	void setErrorHandler(SmartDBErrHandlerFunc func) { errorHandler_ = func; };
	/**
	* NOTICE handler function.
	* The main purpose of the function is to redefine standard behavior of NOTICE.
	* @param proc pointer of user defined function
	* Example:
	* \code
	* struct A
	* {
	* 	static void myNoticeReceiver(const char *msg)
	* 	{
	* 		std::cout << "The message is: " << msg << std::endl;
	* 	}
	* };
	* dbExec->setCustomNoticeReceiver(&A::myNoticeReceiver);
	* \endcode
	*/
	void setCustomNoticeReceiver(const UsrNoticeReceiver proc) const;
	/**
	* Function for removing custom NOTICE handler set before.
	* Example:
	* \code
	* dbExec->clearNoticeReceiver();
	* \endcode
	*/
	void clearNoticeReceiver(void) const;
	template<class T>
	void printRows_(T &tbl, std::ostream&, const char* delim = "|") throw(...);
protected:
	//internal function, should be called only thru the public wrappers
	template <typename LtPred, typename T>
	void sortRows(T &tobj);

	template <typename V> void clearData(V&) const;
	WinLogger* getLogger() const;
	//static CDBExecutor* Instance();
	std::string getTypeName(bool cut = true) const;
	/**
	* Gets number of selected rows
	*
	* @param result query result  
	* @return number of selected rows
	* @remarks 
	*/
	int getRowsCount(PGresult*& result) const;

	CDBExecutor(const CDBExecutor&);
	CDBExecutor& operator=(const CDBExecutor&);

	template<class T>
	void getRows_(Detail::DBManager::QueryResults &results, T &tbl) throw(...);

	template <class T>
	void putRows_(T &tbl, bool prepared, const std::string &stmtName, int format);

	const unsigned getNumPar(const std::string &sqlClause) const;
	template <typename T> T *getTableInstance();
	template<class T> void performSafeInsert(T &tbl, const bool prepared = false, const std::string &stmtName = "");
	template<class T> void performSafeInsert_(T &tbl, const bool prepared, const std::string &prepKey, int format);

	bool isInPrepCmdRep(const std::string &stmtName) const;
	void addToPrepCmdRep(const std::string &cmd, const std::string &stmtName);
	bool getFromPrepCmdRep(/*IN*/const std::string &stmtName, /*OUT*/std::string &prepCmd);
	template <class T> int addArrayParam(PGparam*& param, const char* paramType, const T& v, int i, int j);
	template <class T> int addArrayParam(PGparam*& param, const char* paramType, const boost::numeric::ublas::matrix<T>& v, int i, int j);
	template <> int addArrayParam(PGparam*& param, const char* paramType, const boost::numeric::ublas::matrix<PGvarcharContainer>& v, int i, int j);
	template <class T> int addArrayParam(PGparam*& param, const char* paramType, const boost::numeric::ublas::vector<T>& v, int i, int j);
	template <> int addArrayParam(PGparam*& param, const char* paramType, const boost::numeric::ublas::vector<PGvarcharContainer>& v, int i, int j);
	std::string getSelfAsString(void) const throw();
};

template<class T>
void CDBExecutor::getRows_(Detail::DBManager::QueryResults &results, T &tbl)
{	
	typename T::row_type row_type;

	for(; results.hasMoreRows(); results.nextRow())
	{
		boost::shared_ptr<typename T::row_type> r(new typename T::row_type());

		unsigned int idx = 0; //columns counter
		fusion::for_each(*r.get(), Detail::getCell(results, idx, tbl.colTypes_));

		tbl.push_back(r);
	}
	cleanUp();
};

template <class T>
void CDBExecutor::putRows_(T &tbl, bool prepared, const std::string &prepKey, int format)
{
	//accessing first row
	typename T::internal_iterator i = tbl.begin_internal();

	std::auto_ptr<Detail::DBManager> pgMan( new Detail::DBManager(*connManager_) );

	for(;i != tbl.end_internal(); ++i)
	{
		typename boost::shared_ptr<T::row_type> &row = *i;

		std::auto_ptr<Detail::DBManager::QueryParams> res(pgMan->getQueryParam());

		unsigned int idx = 0; //columns counter
		fusion::for_each(*row.get(), Detail::putCell(*res.get(), idx, tbl.colTypes_));

		if (prepared)
		{
			pgMan->execParamPrepared(*res.get(), prepKey, format);
		}
		else
		{
			pgMan->execParamQuery(*res.get(), stmt_, format);
		}
	}
};

template <class V> 
void CDBExecutor::addParam(const char *paramType, const V* value, const boost::false_type&)//buildin types, will be send by value
{
	params_->put_value(paramType, value);
}
template <class T>
void CDBExecutor::addParam(const char *paramType, multi_array<T> &in_array)
{
	if (!params_.get())
	{
		params_.reset(pg_mng_->getQueryParam());
	}

	unsigned int idx = 0; //columns counter
	//*params.get(), Detail::putCell(*res.get(), idx, colTypes));
	std::vector<std::string> col;
	col.push_back(paramType);

	Detail::putCell pc(*params_.get(), idx, col);

	multi_array<T> *ref = &in_array;

	pc(ref);
}

template <class V> 
void CDBExecutor::addParam(const char *paramType, const V* value, const boost::true_type&)//not buildin types,  will be send by value
{
	params_->put_value(paramType, *value);
}

template <typename T>
void CDBExecutor::addParam(const char *paramType, const T& value)
{
	if (!params_.get())
	{
		params_.reset(pg_mng_->getQueryParam());
	}


	prepareConnection();

	addParam(paramType, &value, boost::mpl::if_<boost::mpl::and_< boost::is_arithmetic<T>, boost::mpl::not_<boost::is_pointer<T>>>, boost::true_type,boost::false_type>::type());//todo refactoring on mpl expresion
}
/*
template <typename V>
void CDBExecutor::addParam(const char *paramType, const V& value)
{
ERROR_HANDLER_START();
prepareConnection();
if (!getParam()) createParam(getParam());

PGTypeReg::ePGTypes type = getPGTypeByStr(paramType);
int r = 1;
try
{
switch(type)
{
case PGTypeReg::eBool:
{
PGbool val = boost::numeric_cast<PGbool>(value);
r = PQputf( getParam(), paramType, val);
}
break;
case PGTypeReg::eInt8:
{
PGint8 val = boost::numeric_cast<PGint8>(value);
r = PQputf( getParam(), paramType, val);
}
break;
case PGTypeReg::eInt4:
{
PGint4 val = boost::numeric_cast<PGint4>(value);
r = PQputf( getParam(), paramType, val);
}
break;
case PGTypeReg::eInt2:
{
PGint2 val = boost::numeric_cast<PGint2>(value);
r = PQputf( getParam(), paramType, val);
}
break;
case PGTypeReg::eFloat8:
{
PGfloat8 val = boost::numeric_cast<PGfloat8>(value);
r = PQputf( getParam(), paramType, val);
}
break;	
case PGTypeReg::eFloat4:
{
PGfloat4 val = boost::numeric_cast<PGfloat4>(value);
r = PQputf( getParam(), paramType, val);
}
break;
case PGTypeReg::eVarchar:
{
PGvarchar val = boost::lexical_cast<PGvarchar>(value);
r = PQputf( getParam(), paramType, val);
}
break;
default:
std::string msg = "Unsupported parameter type:";
msg += paramType;

throw Smartdb::usage_error(msg, SOURCELINE);
}
}
catch(std::bad_cast&)
{
getLogger()->Log(logERROR) << "CDBExecutor::AddParam cast failed";
throw;
}
CHECKEQ(r, 0, SOURCELINE);

getLogger()->Log(logTRACE) << "CDBExecutor::AddParam parameter added successfully";
ERROR_HANDLER_END();
};
*/
template <typename V> 
void CDBExecutor::clearData(V& t) const
{
	int nVectorSize = (int)t.size();
	if(nVectorSize) t.clear();
};
template<class T>
void CDBExecutor::execSQLSet(T &tbl, const std::string &SQLCmd, int format /*= 1 */ )
{
	ERROR_HANDLER_START();
	if (params_.get())
	{
		if (params_->get_num_par() > 0)
		{
			throw Smartdb::usage_error("Current number of added parameters is not 0, use clearParams() before calling execSQLSet", SOURCELINE);
		}
	}
	prepareConnection();

	std::string adjustedSQLCmd;
	adjustSQLCmd(SQLCmd, adjustedSQLCmd);

	setStatement(adjustedSQLCmd);
	performSafeInsert_(tbl, false, "", format); //TODO: remove temp string
	ERROR_HANDLER_END();
}

template<class T>
void CDBExecutor::execSQLGet(T &tbl, const std::string &SQLCmd, int format /*= 1*/ )
{
	ERROR_HANDLER_START();

	std::string adjustedSQLCmd;
	adjustSQLCmd(SQLCmd, adjustedSQLCmd);

	if (params_.get())
	{
		Detail::DBManager::QueryResults &res = pg_mng_->execQuery(*params_.get(), SQLCmd, format, true /*do escape*/);
		params_.reset(NULL);
		logger_->Log(logDEBUG) << adjustedSQLCmd.c_str();
		getRows_(res, tbl);
	}
	else
	{
		Detail::DBManager::QueryResults &res = pg_mng_->execQuery(SQLCmd, format, true /*do escape*/);
		logger_->Log(logDEBUG) << adjustedSQLCmd.c_str();
		getRows_(res, tbl);
	}
	ERROR_HANDLER_END();
}

template<class T>
void CDBExecutor::execPreparedSQLGet(T &tbl, const std::string &prepKey, int format /*= 1 */)
{
	ERROR_HANDLER_START();

	std::string prepCmd;
	bool res = getFromPrepCmdRep(prepKey, prepCmd);
	if (!res)
	{
		throw Smartdb::usage_error(boost::str(boost::format("No prepared command found for \"%1%\" command name ") % prepKey.c_str()), SOURCELINE);
		return;
	}

	setStatement(prepCmd);

	if (params_.get())
	{
		Detail::DBManager::QueryResults &res = pg_mng_->execParamPrepared(*params_.get(), prepKey, format);
		params_.reset(NULL);
		getRows_(res, tbl);
	}
	else
	{
		Detail::DBManager::QueryResults &res = pg_mng_->execPrepared(prepKey, format);
		getRows_(res, tbl);
	}
	logger_->Log(logDEBUG) << "Executing prepared \""<< prepKey << "\" command";
	ERROR_HANDLER_END();
}

template<class T>
void CDBExecutor::execPreparedSQLSet(T &tbl, const std::string &prepKey, int format /*= 1 */)
{
	ERROR_HANDLER_START();
	if (params_.get())
	{
		if (params_->get_num_par() > 0)
		{
			throw Smartdb::usage_error("Current number of added parameters is not 0, use clearParams(), before calling execPreparedSQLSet", SOURCELINE);
		}
	}
	std::string prepCmd;
	bool res = getFromPrepCmdRep(prepKey, prepCmd);
	if (!res)
	{
		throw Smartdb::usage_error(boost::str(boost::format("No prepared command found for \"%1%\" command name ") % prepKey.c_str()), SOURCELINE);
	}

	setStatement(prepCmd);

	performSafeInsert_(tbl, true, prepKey, format);
	logger_->Log(logDEBUG) << "Prepared INSERT \""<< prepKey << "\" executed";
	ERROR_HANDLER_END();
}
template<class T>
void CDBExecutor::performSafeInsert_(T &tbl, const bool prepared, const std::string &prepKey, int format)
{
	const unsigned numFoundParam = getNumPar(getStatement());
	putRows_(tbl, prepared, prepKey, format);
};
template <typename LtPred, typename T>
void CDBExecutor::sortRows(T &b)
{
	ERROR_HANDLER_START();

	getLogger()->Log(logTRACE) << "Sorting the " << typeid(b).name() << " structure.";

	std::sort(b.begin_internal(), b.end_internal(), LtPred());

	ERROR_HANDLER_END();
};
template <typename MLtPred, typename SLtPred, typename MSLtGtEqPred, typename ExternalCnt, typename T, typename S>
void CDBExecutor::linkRows(T &b, S &b1, ExternalCnt &cnt)
{
	ERROR_HANDLER_START();

	typedef typename ExternalCnt::internal_link_cnt internal_link_cnt;

	getLogger()->Log(logTRACE) << "Linking the structures: " << typeid(b).name() << " AND " << typeid(b1).name();

	std::sort(b.begin_internal(), b.end_internal(), MLtPred());

	std::sort(b1.begin_internal(), b1.end_internal(), SLtPred());

	typename T::internal_iterator i = b.begin_internal();
	typename S::internal_iterator j = b1.begin_internal(), tempj;

	MSLtGtEqPred comp;

	if (i != b.end_internal())
	{
		cnt.push_back(std::make_pair(*i, internal_link_cnt()));
	}
	for( ; i != b.end_internal() && j != b1.end_internal(); )
	{
		internal_link_cnt &int_cnt = cnt.back().second;

		if (comp.gt(*i, *j))
		{
			++j;
			continue;
		}
		if (comp.lt(*i, *j))
		{
			++i;
			if (i != b.end_internal())
			{
				cnt.push_back(std::make_pair(*i, internal_link_cnt()));
			}
			continue;
		}
		if (!comp.eq(*i, *j))
		{
			std::stringstream s;
			s << "'gt', 'lt', 'eq' methods from ";
			s << typeid(MSLtGtEqPred).name();
			s << " predicate are not mutual exclusive.";
			throw Smartdb::internal_error(s.str(), SOURCELINE);
		}
		tempj = j;
		do
		{
			int_cnt.push_back(*tempj);
			++tempj; 
		}
		while(tempj != b1.end_internal() && comp.eq(*i, *tempj));
		++i;
		if (i != b.end_internal())
		{
			cnt.push_back(std::make_pair(*i, internal_link_cnt()));
		}
	}
	ERROR_HANDLER_END();
};

template<class T>
void CDBExecutor::printRows_(T &tbl, std::ostream &os, const char* delim)
{	
	typename T::internal_iterator it_ = tbl.begin();
	
	//format a header of the report
	FormatColsMap colFormatMap;
	T::StrVec::iterator coln_it_ = tbl.colNames_.begin();
	unsigned int header_sz_ = 0;
	for (int i = 0; coln_it_ != tbl.colNames_.end(); ++coln_it_, ++i)
	{
		std::size_t sz = (*coln_it_).size();
		++header_sz_ += sz;
		colFormatMap.insert(std::make_pair(i, FormatCols(' ', sz)));
	}
	std::ostream_iterator<std::string> os_it_(os, delim);
	std::string header_(header_sz_, '-');
	std::copy(tbl.colNames_.begin(), tbl.colNames_.end(), os_it_);
	os << std::endl;
	os << header_ << std::endl;

	for(; it_ != tbl.end(); ++it_)
	{
		boost::shared_ptr<typename T::row_type> r(new typename T::row_type());
		r = (*it_);
		unsigned int idx = 0; //columns counter
		//typedef fusion::result_of::size<typename T::row_type>::type TblColNumberSize;
		//TblColNumberSize col_sz;
		fusion::for_each(*r.get(), Detail::printCell(os, idx, tbl.colTypes_, colFormatMap));
		os << std::endl;
	}
};