#include "stdafx.h"

#include "exceptions.h"

namespace Smartdb
{


Smartdb::smartdb_exception::smartdb_exception(const std::string &errPoint) : errPointOfError_(errPoint) 
{
};
Smartdb::smartdb_exception::~smartdb_exception() throw ()
{
};

Smartdb::failure::failure(const std::string &err, const std::string &errPoint) :
																			  smartdb_exception(errPoint),
																			  runtime_error(err)
{
};

Smartdb::protocol_violation::protocol_violation(const std::string &err, const std::string &errPoint) :
																				failure(err, errPoint)
{
};

Smartdb::broken_connection::broken_connection(const std::string &err, const std::string &errPoint) :
																				failure(err, errPoint)
{
};

Smartdb::sql_error::sql_error(const std::string &err, const std::string &errPoint) :
																				failure(err, errPoint)
{
};

Smartdb::sql_error::~sql_error() throw ()
{
};

Smartdb::in_doubt_error::in_doubt_error(const std::string &err, const std::string &errPoint) :
																				failure(err, errPoint)
{
};


Smartdb::internal_error::internal_error(const std::string &err, const std::string &errPoint) :
										logic_error("Smartdb internal error: " + err), smartdb_exception(errPoint)
{
};

/*
Smartdb::usage_error::usage_error(const std::string &err, const std::string &errPoint) :
  logic_error(err), smartdb_exception(errPoint)
{
}
*/
/*
Smartdb::argument_error::argument_error(const std::string &err, const std::string &errPoint) :
  invalid_argument(err, errPoint)
{
}*/
/*

Smartdb::conversion_error::conversion_error(const std::string &err, const std::string &errPoint) :
  domain_error(err), smartdb_exception(errPoint)
{
}


Smartdb::range_error::range_error(const std::string &err, const std::string &errPoint) :
  out_of_range(err), smartdb_exception(errPoint)
{
}
*/
void ThrowSQLErrorByCode(const char* code, const std::string &err, const std::string &errPoint)
{
	if (code) 
		switch (code[0])
		{
		case '0':
			switch (code[1])
			{
				case '8':
				{
					if (!strcmp(code, "08P01"))
					throw protocol_violation(err, errPoint);
					throw broken_connection(err, errPoint);
				}
				case 'A':
				{
					throw feature_not_supported(err, errPoint);
				}
			}
			break;
		case '2':
			switch (code[1])
			{
			case '2':
				throw data_exception(err, errPoint);
			case '3':
				if (strcmp(code,"23001")==0) throw restrict_violation(err, errPoint);
				if (strcmp(code,"23502")==0) throw not_null_violation(err, errPoint);
				if (strcmp(code,"23503")==0) throw foreign_key_violation(err, errPoint);
				if (strcmp(code,"23505")==0) throw unique_violation(err, errPoint);
				if (strcmp(code,"23514")==0) throw check_violation(err, errPoint);
				throw integrity_constraint_violation(err, errPoint);
			case '4':
				throw invalid_cursor_state(err, errPoint);
			case '6':
				throw invalid_sql_statement_name(err, errPoint);
			}
			break;
		case '3':
			switch (code[1])
			{
			case '4':
				throw invalid_cursor_name(err, errPoint);
			case 'F':
				throw invalid_schema_name(err, errPoint);
			}
			break;
		case '4':
			switch (code[1])
			{
			case '2':
			if (strcmp(code,"42501")==0) throw insufficient_privilege(err, errPoint);
			if (strcmp(code,"42601")==0) throw syntax_error(err, errPoint);
			if (strcmp(code,"42703")==0) throw undefined_column(err, errPoint);
			if (strcmp(code,"42883")==0) throw undefined_function(err, errPoint);
			if (strcmp(code,"42P01")==0) throw undefined_table(err, errPoint);
			if (strcmp(code,"42804")==0) throw datatype_mismatch(err, errPoint);
			}
			break;
		case '5':
			switch (code[1])
			{
			case '3':
				if (strcmp(code,"53100")==0) throw disk_full(err, errPoint);
				if (strcmp(code,"53200")==0) throw out_of_memory(err, errPoint);
				if (strcmp(code,"53300")==0) throw too_many_connections(err, errPoint);
				throw insufficient_resources(err, errPoint);
			}
		}
	std::stringstream ss;
	ss << code;
	ss << " has the following message: " << err;
	throw internal_error("Unrecognized sql response code " + ss.str(), errPoint);
}
void ThrowSQLError(const PGresult *res, const std::string &err, const std::string &errPoint)
{
	assert(res != NULL);
	// Try to establish more precise error type, and throw corresponding exception
	const char *const code = PQresultErrorField(res, PG_DIAG_SQLSTATE);
	ThrowSQLErrorByCode(code, err, errPoint);
}
void ThrowSQLError(const char *code, const std::string &err, const std::string &errPoint)
{
	assert(code != NULL);
	// Try to establish more precise error type, and throw corresponding exception
	ThrowSQLErrorByCode(code, err, errPoint);
}
void CheckStatus(PGconn *conn, const PGresult *res, const std::string &errPoint)
{	
	//[Added by Nikolay]: PQgetErrorField - libpqtypes function extracts error info saved by pqt_setresultfields 
	//pqt_setresultfields - each time being called it cleans up the PGresult which is absolutely needed for appropriate status checking
	//see more details at exec.c
	const char *lpqtypes_errcode = LibPQExp::_PQgetErrorField(PG_DIAG_SQLSTATE);
	if (!res && !lpqtypes_errcode)
	{
		const char *msg = PQerrorMessage(conn);
		if (!msg)
		{
			throw Smartdb::internal_error(msg, errPoint);
		}
		else
		{
			throw Smartdb::failure(msg, errPoint);
		}
	}
	std::string Err;

	switch (PQresultStatus(res))
	{
		case PGRES_EMPTY_QUERY: // The string sent to the backend was empty.
		case PGRES_COMMAND_OK: // Successful completion of a command returning no data
		case PGRES_TUPLES_OK: // The query successfully executed
			break;

		case PGRES_COPY_OUT: // Copy Out (from server) data transfer started
		case PGRES_COPY_IN: // Copy In (to server) data transfer started
			break;

		case PGRES_BAD_RESPONSE: // The server's response was not understood
		case PGRES_NONFATAL_ERROR:
		case PGRES_FATAL_ERROR:
			//getting a message either from libpq or libpqtypes function 
			Err = lpqtypes_errcode ? LibPQExp::_PQgetErrorField(PG_DIAG_MESSAGE_PRIMARY) : PQresultErrorMessage(res);
			//if we have got sqlcode in lpqtypes_errcode then let's try to use it
			if (lpqtypes_errcode)
				ThrowSQLError(lpqtypes_errcode, Err, errPoint);
			else 
				ThrowSQLError(res, Err, errPoint);
			break;
		default:
		{
			std::stringstream ss;
			ss << int(PQresultStatus(res));
			throw internal_error("Unrecognized response code " + ss.str(), errPoint);
		}
	}
}

}; //end of Smartdb namespace
