#pragma once

#include <string>
#include <fstream>
#include <iostream>
#include <stack>
#include <boost/shared_ptr.hpp>//used for shared_ptr

enum TLogLevel {logFATAL, logERROR, logWARNING, logINFO, logDEBUG, logTRACE};

namespace NsSLogger
{
	template <typename TimeReporter>
	class SLogger
	{
	public:
		virtual ~SLogger();
		std::ofstream& Log(TLogLevel level);
		
		void PushCachedLog();
		void PopCachedLog();
		std::string GetTopCachedLog() { return m_cachedMsg.str(); };
		void ClearTopCachedLog() { m_cachedMsg.str(""); };

		std::stringstream& LogToCache();

		static boost::shared_ptr<SLogger<TimeReporter> > 
			Init(const std::string &logFilePath, 
				const bool clrLogFile = true, 
				const std::string &strReportLevel = "",
				const std::string &defaultFileName = "");

		TLogLevel GetReportingLevel() { return m_reportingLevel; } ;
		void LogCachedMessage(TLogLevel level);
		void SetIgnoreFirstDelimiter() { m_ignoreFirstDelimiter = true; }
		void ResetIgnoreFirstDelimiter() { m_ignoreFirstDelimiter = false; }
		void Flush() { m_outputFileStream.flush(); }

		void SetReportingLevel(TLogLevel val);
	protected:
		void OpenFile(const std::string &filePath, const bool clrLogFile, const std::string &defaultFileName);
		static std::string ToString(TLogLevel level);
		static TLogLevel FromString(const std::string& level);
		
	protected:
		std::ofstream m_outputFileStream, m_nullFileStream;
		TLogLevel m_reportingLevel;
		std::stringstream m_cachedMsg;
		
		std::stack<std::string> m_cachedLogStack;

		bool m_ignoreFirstDelimiter;

	private:
		SLogger();
		SLogger(const SLogger&);
		SLogger& operator =(const SLogger&);
	};

	template <typename T>
	SLogger<T>::SLogger()
		: m_nullFileStream("NUL"),
		m_ignoreFirstDelimiter(false)
	{
	}

	template <typename TimeReporter>
	std::ofstream& SLogger<TimeReporter>::Log(TLogLevel level)
	{
		if (level > GetReportingLevel())
			return m_nullFileStream;
		m_outputFileStream << std::endl;
		m_outputFileStream << "- " << TimeReporter::NowTime();
		m_outputFileStream << " " << ToString(level) << ": ";
		return m_outputFileStream;
	}
	template <typename TimeReporter>
	void SLogger<TimeReporter>::PushCachedLog()
	{
		m_cachedLogStack.push(m_cachedMsg.str());
		m_cachedMsg.str("");
	}
	template <typename TimeReporter>
	void SLogger<TimeReporter>::PopCachedLog()
	{
		if (!m_cachedLogStack.empty())
		{
			m_cachedMsg.str("");
			m_cachedMsg  << m_cachedLogStack.top();
			m_cachedLogStack.pop();
		}
	}
	template <typename TimeReporter>
	std::string SLogger<TimeReporter>::ToString(TLogLevel level)
	{
		static const char* const buffer[] = {"FATAL", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE"};
		return buffer[level];
	}
	template <typename TimeReporter>
	std::stringstream& SLogger<TimeReporter>::LogToCache()
	{
		return m_cachedMsg;
	}
	template <typename TimeReporter>
	void SLogger<TimeReporter>::LogCachedMessage(TLogLevel level)
	{
		if (level <= GetReportingLevel())
		{	
			m_outputFileStream << std::endl;
			m_outputFileStream << "- " << TimeReporter::NowTime();
			m_outputFileStream << " " << ToString(level) << ": ";
			m_outputFileStream << m_cachedMsg.str();
			Flush();
		}
		m_cachedMsg.str("");
	}
	template <typename T>
	SLogger<T>::~SLogger()
	{
		m_outputFileStream.close();
	}

	template <typename TimeReporter>
	boost::shared_ptr<SLogger<TimeReporter> >
	SLogger<TimeReporter>::Init(const std::string &logFilePath, const bool clrLogFile, const std::string &strReportLevel, const std::string &defaultFileName)
	{
		boost::shared_ptr<SLogger<TimeReporter> > logger(new SLogger<TimeReporter>());
		logger->OpenFile(logFilePath, clrLogFile, defaultFileName);
		const TLogLevel reportLevel = FromString(strReportLevel);
		logger->SetReportingLevel(reportLevel);
		return logger;
	}
	
	template <typename TimeReporter>
	void SLogger<TimeReporter>::OpenFile(const std::string &filePath, const bool clrLogFile, const std::string &defaultFileName)
	{
		std::ios_base::openmode openMode = std::ios::out;
		if (clrLogFile)
			openMode |= std::ios::trunc;
		else
			openMode |= std::ios::app;

		m_outputFileStream.open(filePath.c_str(), openMode);
		
		if (m_outputFileStream.fail())
		{
			bool error = true;
			if (defaultFileName.length() > 0)
			{
				m_outputFileStream.close();
				m_outputFileStream.clear();

				//try to create default log
				m_outputFileStream.open(defaultFileName.c_str(), openMode);
				if (!m_outputFileStream.fail())
				{
					error = false;
				}
			}
			if (error)
			{
				m_outputFileStream.close();
				throw std::runtime_error("Cannot open file for logger");	
			}
		}
	}

	template <typename TimeReporter>
	void SLogger<TimeReporter>::SetReportingLevel(TLogLevel val)
	{
		m_reportingLevel = val;
	}
	template <typename TimeReporter>
	TLogLevel SLogger<TimeReporter>::FromString(const std::string& level)
	{
		if (!level.compare("FATAL"))
			return logFATAL;
		else
		if (!level.compare("ERROR"))
			return logERROR;
		else
		if (!level.compare("WARNING"))
			return logWARNING;
		else
		if (!level.compare("INFO"))
			return logINFO;
		else
		if (!level.compare("DEBUG"))
			return logDEBUG;
		return logTRACE;
	}
};